from logging import getLogger
from typing import Tuple, Dict, Optional, Union, List, Callable, Any

from src.surrealist.errors import OperationOnClosedConnectionError
from src.surrealist.utils import SurrealResult, DEFAULT_TIMEOUT

logger = getLogger("connection")


def connected(func):
    """
    Decorator for methods to make sure underlying connection is alive (connected to DB)

    :param func: method to decorate
    :raise OperationOnClosedConnectionError: if connection is already closed
    """

    def wrapped(*args, **kwargs):
        # args[0] is a self argument in methods
        if not args[0].is_connected():
            message = "Your connection already closed"
            logger.error(message, exc_info=False)
            raise OperationOnClosedConnectionError(message)
        return func(*args, **kwargs)

    return wrapped


class Connection:
    """
    Parent for connection objects, contains all public methods to work with API
    """

    def __init__(self, db_params: Optional[Dict] = None, credentials: Tuple[str, str] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        """
        Init any connection to use
        :param db_params: optional parameter, if it is not None, should be like {"NS": "test", "DB": "test"}
        :param credentials: optional pair of user and pass for auth, like ("root", "root")
        :param timeout: timeout in seconds to wait connection results and responses
        """
        self._db_params = db_params
        self._credentials = credentials
        self._connected = False
        self._timeout = timeout

    def close(self):
        """
        Closes the connection. You can not and should not use connection object after that
        """
        logger.info("Connection was closed")
        self._connected = False

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Checks the connection is still alive and usable

        :return: True if connection is usable, False otherwise
        """
        return self._connected

    def use(self, namespace: str, database: str) -> SurrealResult:
        return NotImplemented

    def info(self) -> SurrealResult:
        return NotImplemented

    def authenticate(self, token: str) -> SurrealResult:
        return NotImplemented

    def invalidate(self) -> SurrealResult:
        return NotImplemented

    def let(self, name: str, value) -> SurrealResult:
        return NotImplemented

    def unset(self, name: str) -> SurrealResult:
        return NotImplemented

    def live(self, table_name: str, callback: Callable[[Dict], Any], return_diff: bool = False) -> SurrealResult:
        return NotImplemented

    def kill(self, live_query_id: str) -> SurrealResult:
        return NotImplemented

    def signup(self, namespace: str, database: str, scope: str, params: Optional[Dict] = None) -> SurrealResult:
        return NotImplemented

    def signin(self, user: str, password: str, namespace: Optional[str] = None,
               database: Optional[str] = None, scope: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def select(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def create(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def insert(self, table_name: str, data: Union[Dict, List]) -> SurrealResult:
        return NotImplemented

    def update(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def merge(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def delete(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        return NotImplemented

    def patch(self, table_name: str, data: Union[Dict, List], record_id: Optional[str] = None,
              return_diff: bool = False) -> SurrealResult:
        return NotImplemented

    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        return NotImplemented

    def import_data(self, path) -> SurrealResult:
        return NotImplemented

    def export(self) -> str:
        return NotImplemented

    def ml_import(self, path) -> SurrealResult:
        return NotImplemented

    def ml_export(self, name: str, version: str) -> str:
        return NotImplemented



import base64
import json
import urllib.parse
import urllib.request
from http.client import HTTPResponse, RemoteDisconnected
from logging import getLogger
from typing import Optional, Tuple, Dict, Union
from urllib.error import URLError, HTTPError

from src.surrealist.errors import HttpClientError, TooManyNestedLevelsError
from src.surrealist.utils import ENCODING, DEFAULT_TIMEOUT, crop_data, mask_pass

logger = getLogger("http_client")


class HttpClient:
    """
    Http-client for working with http endpoints and abilities of SurrealDB
    """

    def __init__(self, base_url: str, headers: Optional[Dict] = None, credentials: Optional[Tuple[str, str]] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        self._base_url = base_url
        self._credentials = credentials
        self._timeout = timeout
        headers = headers or {}
        self._headers = {"Accept": "application/json", "User-Agent": "py_surreal http-client", **headers}
        if credentials:
            self._user, self._pass = credentials
            base64string = base64.encodebytes(f'{self._user}:{self._pass}'.encode(ENCODING))[:-1]
            self._headers["Authorization"] = f"Basic {base64string.decode(ENCODING)}"

    def get(self, path: str = '') -> HTTPResponse:
        """
        Represents GET method
        :param path: endpoint to request
        :return: response
        """
        return self.request("GET", None, path)

    def post(self, data: Dict, path: str = '') -> HTTPResponse:
        """
        Represents POST method
        :param data: json or bytes data
        :param path: endpoint to request
        :return: response
        """
        return self.request("POST", data, path)

    def put(self, data: Dict, path: str = '') -> HTTPResponse:
        """
        Represents POST method
        :param data: json or bytes data
        :param path: endpoint to request
        :return: response
        """
        return self.request("PUT", data, path)

    def patch(self, data: Dict, path: str = '') -> HTTPResponse:
        """
        Represents POST method
        :param data: json or bytes data
        :param path: endpoint to request
        :return: response
        """
        return self.request("PATCH", data, path)

    def request(self, method: str, data: Optional[Union[Dict, str]], path: str = '',
                not_json: bool = False) -> HTTPResponse:
        response = None
        url = f'{self._base_url}{path}'
        options = {'method': method, 'headers': self._headers}
        if method not in ("GET", "DELETE"):
            try:
                js = json.dumps(data).encode(ENCODING) if not not_json else data.encode(ENCODING)
            except RecursionError as e:
                logger.error("Cant serialize object, too many nested levels")
                raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
            options['data'] = js
        try:
            req = urllib.request.Request(url, **options)
            logger.debug("Request to %s, options: %s, timeout: %d", url, mask_opts(options), self._timeout)
            response = urllib.request.urlopen(req, timeout=self._timeout)
            return response
        except HTTPError as e:
            logger.error("Http error on request %s, info: %s", url, e)
            if response:
                response.close()
            # we need it to check status code and text on error
            return e
        except (URLError, RemoteDisconnected, ConnectionResetError) as e:
            if response:
                response.close()
            logger.error("Error on connecting to %s, info: %s", url, e)
            raise HttpClientError(f"Error on connecting to '{url}'") from e


def mask_opts(options: Dict) -> Dict:
    """
    Hide authorization data with asterisks, so "Authorization": "Basic ***" will be on return

    :param options: dict with options(headers)
    :return: same dict or dict with hided auth data
    """
    masked_opts = {}
    for key, value in options.items():
        if key == "headers" and "Authorization" in value:
            value = {**value, "Authorization": "Basic ******"}
        elif key == "data":
            value = crop_data(mask_pass(str(value)))
        masked_opts[key] = value
    return masked_opts



from .surreal import Surreal
from .connections import WebSocketConnection, HttpConnection
from .errors import *
from .utils import DATA_LENGTH_FOR_LOGS

__version__ = "0.1.1"

__all__ = ("Surreal", "WebSocketConnection", "HttpConnection", "PySurrealError", "HttpConnectionError",
           "HttpClientError", "SurrealConnectionError", "WebSocketConnectionError", "WebSocketConnectionClosedError",
           "ConnectionParametersError", "CompatibilityError", "OperationOnClosedConnectionError",
           "DATA_LENGTH_FOR_LOGS")



class PySurrealError(Exception):
    """
    Parent for all types of errors in py_surreal, so you can use more broad error in except clauses

    try:
        // some job with connection
    except PySurrealError:
        // do something on error
    """
    # TODO make code pretty-formatted


class SurrealConnectionError(PySurrealError):
    """
    Raises on any problem with connection to SurrealDB, check url and other parameters
    """


class HttpClientError(PySurrealError):
    """
    Raises on http-client specific errors, in normal situations user should never see this
    """


class HttpConnectionError(PySurrealError):
    """
    Raises on troubles with connection or authentication, when using http-connection
    """


class WebSocketConnectionError(PySurrealError):
    """
    Raises on troubles with connection or data, when using websocket-connection
    """


class WebSocketConnectionClosedError(PySurrealError):
    """
    Raises only if websocket connection was closed, but one or more websocket clients still alive
    (waiting for some data)
    """


class ConnectionParametersError(PySurrealError):
    """
    Raises only if websocket connection gets wrong namespace/database parameters, in normal situations user
    should never see this
    """


class OperationOnClosedConnectionError(PySurrealError):
    """
    Raises on attempt to use already closed connection

    """


class CompatibilityError(PySurrealError):
    """
    Raises on attempt to use methods incompatible for that transport (websocket ot http), for example **live** do not
    work on http-connection.
    More on this here: url
    """
    # TODO link here


class TooManyNestedLevelsError(PySurrealError):
    """
    Raises when str or json cant handle the object because of too deep nesting and recursion limit in python
    See documentation: url
    """
    # TODO link here



import urllib.parse
from logging import getLogger, ERROR, DEBUG, basicConfig, INFO
from typing import Tuple, Optional

from src.surrealist.clients import HttpClient
from src.surrealist.connections.connection import Connection
from src.surrealist.connections.http_connection import HttpConnection
from src.surrealist.connections.ws_connection import WebSocketConnection
from src.surrealist.utils import DEFAULT_TIMEOUT, _set_length, DATA_LENGTH_FOR_LOGS, ENCODING, OK, HTTP_OK

FORMAT = '%(asctime)s : %(threadName)s : %(name)s : %(levelname)s : %(message)s'
basicConfig(level=ERROR, format=FORMAT)
logger = getLogger()


class Surreal:
    """
    Represents the SurrealDB client, all connections should be established via this class. By default, connection will
    use websocket transport (recommended). This class use log level to monitor all actions, can check version and state
    of the SurrealDb server
    """

    def __init__(self, url: str, namespace: Optional[str] = None, database: Optional[str] = None,
                 credentials: Tuple[str, str] = None, use_http: bool = False, timeout: int = DEFAULT_TIMEOUT,
                 log_level: str = "ERROR"):
        """
        Initiating all parameters for connection, this method do not check or validate anything by itself, just save
        data for future use. To make sure your url is valid and accessible -use is_ready method of Surreal object.


        :param url: database url, for local database http://127.0.0.1:8000/
        :param namespace: namespace to use
        :param database: database to use, make sure to use both namespace and database or none of them
        :param credentials: pair of user and password, for example ("root", "root")
        :param use_http: boolean flag of using http transport. Will use websocket-client if False.
        It is strongly recommended to use websocket transport as it is more powerful.
        :param timeout: connection timeout in seconds
        :param log_level: level of logging for debug purposes, one of (DEBUG, INFO, ERROR) allowed, where ERROR
        is default, INFO - to see only operations, DEBUG - to see all, including transport requests/responses
        """
        self.log_level = log_level
        self.log_data_length = DATA_LENGTH_FOR_LOGS
        self.set_log_level(log_level)
        self.client = HttpConnection if use_http else WebSocketConnection
        self.db_params = {}
        if namespace:
            self.db_params["NS"] = namespace
        if database:
            self.db_params["DB"] = database
        if not self.db_params:
            self.db_params = None
        self._url = url
        self._is_http_url = url.startswith("http")
        # we should change only http endpoints
        if self._is_http_url and not url.endswith("/"):
            self._url= f"{url}/"
        self.credentials = credentials
        self.timeout = timeout
        self._possible_url = self._url
        # we can try to predict http url when wss url was specified
        if not self._is_http_url:
            _url = urllib.parse.urlparse(url.lower())
            self._possible_url = f"{_url.scheme.replace('ws', 'http')}://{_url.netloc}/"
            logger.info("Predicted url is: %s", self._possible_url)

    def set_log_length_for_data(self, length: int):
        """
        Setting maximum string length for object(json) in logs, it is not always desirable to see all data that comes
        in and out in logs, because some objects(jsons) can be very large, but for debug purposes you can increase it
        to see full picture. However, is not recommended to set length too small for speed reasons, or set it too large,
        except when you in the debug mode. Default is 300 chars, which is enough for standard SurrealDB messages.

        :param length: new maximum length for one string in logs
        :return: None
        """
        _set_length(length)
        self.log_data_length = length

    def set_log_level(self, level: str):
        """
        Setting log level for all underlying connections of that Surreal object.
        There are 3 options:
        - ERROR which is default and recommended for most situations, you will see only errors
        - INFO which is show all information about connection operations
        - DEBUG, when you will see all logs, including requests and responses via transport level. This level is very
        useful for debugging and finding some useful information in logs

        Notice: authorization params and passwords cant be seen in logs in any time, if you see it, please report
        an issue
        Notice: you will see logs of this library(in and out), but not SurrealDb server logs

        :param level: one of "DEBUG", "INFO", "ERROR"
        :return: None
        """
        if level not in ("DEBUG", "INFO", "ERROR"):
            raise ValueError("Log level shoud be one of (DEBUG, INFO, ERROR), where ERROR is default")
        levels = {"DEBUG": DEBUG, "INFO": INFO, "ERROR": ERROR}
        new_level = levels[level]
        logger.setLevel(new_level)
        logger.debug("Log level switch to %s", level)
        self.log_level = level

    def connect(self) -> Connection:
        """
        Actually connects to database via chosen transport (websocket or http), uses specified namespace, database and
        credentials parameters, so can raise exception if connect will fail. If this method succeeded - you are
        ready to go with SurrealDB

        :return: connection object to work with SurrealDB
        :raise SurrealConnectionError: if cant connect with specified parameters
        """
        return self.client(self._url, db_params=self.db_params, credentials=self.credentials, timeout=self.timeout)

    def is_ready(self) -> bool:
        """
        Checks that SurrealDB server is up and running. Under the hood it calls **health** and **status** methods to
        make sure database web server is running and database server and storage engine are running.

        Notice: Method uses the specified url, but not namespace, database or credentials, so success of this method
        (return True) should not be interpreted as authorization success

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#status

        https://docs.surrealdb.com/docs/integration/http#health

        :return: True if you can use this server, False otherwise
        """
        return self.status() == OK and self.health() == OK

    def status(self) -> str:
        """
        Checks that SurrealDB web server is running.

        Refer to: https://docs.surrealdb.com/docs/integration/http#status

        :return: "OK" if everything is ok, or error text
        """
        code, text = self.__get("status")
        return OK if code == HTTP_OK else text

    def health(self) -> str:
        """
        Checks that SurrealDB server and storage engine are running.

        Refer to: https://docs.surrealdb.com/docs/integration/http#health

        :return: "OK" if everything is ok, or error text
        """
        code, text = self.__get("health")
        return OK if code == HTTP_OK else text

    def version(self) -> str:
        """
        Get SurrealDB version

        Refer to: https://docs.surrealdb.com/docs/integration/http#version

        :return: version, for example "surrealdb-1.1.1"
        """
        return self.__get("version")[1]

    def __get(self, endpoint: str) -> Tuple[int, str]:
        with HttpClient(self._possible_url, timeout=DEFAULT_TIMEOUT).get(endpoint) as resp:
            status, text = resp.status, resp.read().decode(ENCODING)
            body = "is empty" if not text else text
            logger.info("Response from /%s, status_code: %s, body: %s", endpoint, status, body)
            if status != HTTP_OK:
                logger.error("Status is %s for predicted http %s", status, self._possible_url)
            return status, text



import json
from logging import getLogger
from pathlib import Path
from typing import Tuple, Dict, Optional, Union, Any

from src.surrealist.clients.http_client import HttpClient
from src.surrealist.connections.connection import Connection, connected
from src.surrealist.errors import (SurrealConnectionError, HttpClientError, CompatibilityError, HttpConnectionError,
                                   TooManyNestedLevelsError)
from src.surrealist.utils import (ENCODING, to_result, SurrealResult, DEFAULT_TIMEOUT, crop_data, mask_pass, HTTP_OK)

logger = getLogger("http_connection")


class HttpConnection(Connection):
    """
    Represents http transport and abilities to work with SurrealDb. It is not recommended connection, use websocket on
    any doubt, this connection exists for compatibility reasons. Some features as live query is not possible on this
    transport, but import and export are.
    Refer to py_surreal documentation and here https://docs.surrealdb.com/docs/integration/http

    Each method call creates new http short-live connection.

    On creating, this object tries to create connection with specified data and will raise exception on fail.
    """

    def __init__(self, url: str, db_params: Optional[Dict] = None, credentials: Tuple[str, str] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        super().__init__(db_params, credentials, timeout)
        self._url = url
        self._http_client = HttpClient(url, headers=db_params, credentials=credentials, timeout=timeout)
        try:
            is_ready = self._is_ready()
        except HttpClientError:
            is_ready = False
        if not is_ready:
            logger.error("Cant connect to %s OR /status and /health are not OK", url)
            raise SurrealConnectionError(f"Cant connect to {url} OR /status and /health are not OK.\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")
        self._connected = True
        masked_creds = None if not credentials else (credentials[0], "******")
        logger.info("Connected to %s, params: %s, credentials: %s, timeout: %s", url, db_params, masked_creds, timeout)

    def _is_ready(self) -> bool:
        return self._simple_get("status")[0] == HTTP_OK and self._simple_get("health")[0] == HTTP_OK

    @connected
    def signin(self, user: str, password: str, namespace: Optional[str] = None, database: Optional[str] = None,
               scope: Optional[str] = None) -> SurrealResult:
        """
        This method allows you to sign in a root, namespace, database or scope user against SurrealDB

        Refer to: https://docs.surrealdb.com/docs/integration/http#signin

        Example:
        http_connection.signin('root', 'root') # sign in as root user
        http_connection.signin('root', 'root', namespace='test', database='test') # sign in as db user


        :param user: name of the user
        :param password: user password
        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :param scope: name of the scope to use
        :return: result of request
        """
        opts = {"user": user, "pass": password}
        if namespace:
            opts['ns'] = namespace
        if database:
            opts['db'] = database
        if scope:
            opts['scope'] = scope
        logger.info("Operation: SIGNIN. Data: %s", crop_data(mask_pass(str(opts))))
        text = raise_if_not_http_ok(self._simple_request("POST", "signin", opts))
        return to_result(text)

    @connected
    def signup(self, namespace: str, database: str, scope: str, params: Optional[Dict] = None) -> SurrealResult:
        """
        This method allows you to sign up a user

        Refer to: https://docs.surrealdb.com/docs/integration/http#signup

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

        Example:
        http_connection.signup(namespace='test', database='test', scope='user_scope',
                              params={'user': 'john:doe', 'pass': '123456'})

        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :param scope: name of the scope to use
        :param params: dict with user and pass, for example {"user":"root", "pass":"root"}
        :return: result of request
        """
        params = params or {}
        body = {"ns": namespace, "db": database, "sc": scope, **params}
        logger.info("Operation: SIGNUP. Data: %s", crop_data(mask_pass(str(body))))
        text = raise_if_not_http_ok(self._simple_request("POST", "signup", body))
        return to_result(text)

    @connected
    def select(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method selects either all records in a table or a single record

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#get-table
        https://docs.surrealdb.com/docs/integration/http#get-record

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

        Examples:
        http_connection.select("article") # select all records in article table
        http_connection.select("article:first") # select one article with id 'first'
        http_connection.select("article", "first") # select one article with id 'first', analog of previous

        Notice: do not specify id twice: in table name and in record_id, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to select
        :param record_id: optional parameter, if exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        url = f"key/{table_name}" if record_id is None else f"key/{table_name}/{record_id}"
        logger.info("Operation: SELECT. Path: %s", crop_data(url))
        text = raise_if_not_http_ok(self._simple_get(url))
        return to_result(text)

    @connected
    def create(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method creates a record either with a random or specified record_id. If no id specified in record_id or
        in data arguments, then id will be generated by SurrealDB

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#post-table
        https://docs.surrealdb.com/docs/integration/http#post-record

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/create

        Examples:
        http_connection.create("person", {"name": "John Doe"}) # create one record in person table with random id
        http_connection.create("person", {"id":"my_id", "name": "John Doe"}) # create one record in person table
        with specified id
        http_connection.create("person", {"name": "John Doe"}, "my_id") # create one record in person table
        with specified id
        http_connection.create("person:my_id", {"name": "John Doe"}) # create one record in person table
        with specified id

        Notice: do not specify id twice, for example in table name and in data, it will cause error on SurrealDB side


        :param table_name: table name or table name with record_id to create
        :param data: dict with data to create
        :param record_id: optional parameter, if exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        url = f"key/{table_name}" if record_id is None else f"key/{table_name}/{record_id}"
        logger.info("Operation: CREATE. Path: %s, data: %s", crop_data(url), crop_data(str(data)))
        text = raise_if_not_http_ok(self._simple_request("POST", url, data))
        return to_result(text)

    @connected
    def update(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method can be used to update or modify records in the database. So all old fields will be deleted and new
        will be added, if you wand just to add field to record, keeping old ones - use merge method instead

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#put-table
        https://docs.surrealdb.com/docs/integration/http#put-record

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

        Example:
        http_connection.update("person:my_id", {"name": "Alex Doe"}) # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted
        http_connection.update("person", {"name": "Alex Doe"}, "my_id") # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted

        Notice: do not specify id twice, for example in table name and in data, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to update
        :param data: dict with data to create
        :param record_id: optional parameter, if exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        url = f"key/{table_name}" if record_id is None else f"key/{table_name}/{record_id}"
        logger.info("Operation: UPDATE. Path: %s, data: %s", crop_data(url), crop_data(str(data)))
        text = raise_if_not_http_ok(self._simple_request("PUT", url, data))
        return to_result(text)

    @connected
    def delete(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method deletes all records in a table or a single record, be careful and do not forget to specify id if you
        do not want to delete all records. This method do not remove table itself, only records in it

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#delete-table
        https://docs.surrealdb.com/docs/integration/http#delete-record

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/delete

        Examples:
        http_connection.delete("person:my_id") # deletes one record in person table
        http_connection.delete("person", "my_id") # deletes one record in person table
        http_connection.delete("person") # deletes all records in person table

        :param table_name: table name or table name with record_id to delete
        :param record_id: optional parameter, if exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        url = f"key/{table_name}" if record_id is None else f"key/{table_name}/{record_id}"
        logger.info("Operation: DELETE. Path: %s", crop_data(url))
        text = raise_if_not_http_ok(self._simple_request("DELETE", url, {}))
        return to_result(text)

    @connected
    def merge(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method merges specified data into either all records in a table or a single record. Old data in records
        will not be deleted, if you want to replace old data with new - use update method

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#patch-table
        https://docs.surrealdb.com/docs/integration/http#patch-record

        Examples:
        http_connection.merge("person",{"active": True}) # "active" will be added to all records in person table
        http_connection.merge("person:my_id", {"active": True}) # "active" will be added to one record in person
        table with specified id

        :param table_name: table name or table name with record_id to merge
        :param data: dict with data to add
        :param record_id: optional parameter, if exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        url = f"key/{table_name}" if record_id is None else f"key/{table_name}/{record_id}"
        logger.info("Operation: PATCH. Path: %s, data: %s", crop_data(url), crop_data(str(data)))
        text = raise_if_not_http_ok(self._simple_request("PATCH", url, data))
        return to_result(text)

    @connected
    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        """
        This method used for execute a custom SurrealQL query

        Attention! Http transport ignores variables, embed them into your query or switch to websocket connection

        Refer to: https://docs.surrealdb.com/docs/integration/http#sql

        For SurrealQL refer to: https://docs.surrealdb.com/docs/surrealql/overview

        Example:
        http_connection.query("SELECT * FROM article;") # gets all records from article table

        :param query: any SurrealQL query to execute
        :param variables: a set of variables used by the query
        :return: result of request
        """
        if variables is not None:
            logger.warning("Variables parameter cant be used on QUERY with http-connection, "
                           "embed them into your query or switch to websocket connection")
        logger.info("Operation: QUERY. Query: %s", crop_data(query))
        text = raise_if_not_http_ok(self._simple_request("POST", "sql", query, not_json=True))
        return to_result(text)

    @connected
    def import_data(self, path: Union[str, Path]) -> SurrealResult:
        """
        This method imports a SurrealQL script file into a local or remote SurrealDB database server.

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#import
        https://docs.surrealdb.com/docs/cli/import

        Refer to: https://docs.surrealdb.com/docs/integration/http#import
        :param path: path to file for import
        :return: result of request
        """
        with open(path, 'rb') as file:
            logger.info("Operation: IMPORT. Path: %s", crop_data(str(path)))
            text = raise_if_not_http_ok(self._simple_request("POST", "import", file.read().decode(ENCODING),
                                                             not_json=True))
        return to_result(text)

    @connected
    def export(self) -> str:
        """
        This method exports all data for a specific namespace and database.

        Refer to: https://docs.surrealdb.com/docs/integration/http#export

        Refer to: https://docs.surrealdb.com/docs/cli/export/


        :return: text of the exported file to use or save
        """
        text = raise_if_not_http_ok(self._simple_get("export"))
        return text

    @connected
    def ml_import(self, path: Union[str, Path]) -> SurrealResult:
        """
        This method imports a SurrealQL ML file into a local or remote SurrealDB database server.

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#ml-import
        https://docs.surrealdb.com/docs/cli/ml/import

        :param path: path to file for import
        :return: result of request
        """
        with open(path, 'rb') as file:
            logger.info("Operation: ML IMPORT. Path: %s", crop_data(str(path)))
            text = raise_if_not_http_ok(self._simple_request("POST", "ml/import", file.read().decode(ENCODING),
                                                             not_json=True))
        return to_result(text)

    @connected
    def ml_export(self, name: str, version: str) -> str:
        """
        This method exports a SurrealML machine learning model from a specific namespace and database. As machine
        learning files can be large, the endpoint outputs a chunked HTTP response.

        Refer to: https://docs.surrealdb.com/docs/integration/http#ml-export

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export

        Example:
        http_connection.ml_export("prediction", "1.0.0")

        :param name: name of ML model
        :param version: version of ML model
        :return:  text of the exported file to use or save
        """
        logger.info("Operation: ML EXPORT. Name %s Version %s", name, version)
        text = raise_if_not_http_ok(self._simple_get(f"ml/export/{name}/{version}"))
        return text

    @connected
    def let(self, name: str, value: Any) -> SurrealResult:
        """
        This method sets and stores a value which can then be used in a subsequent query

        Http transport use **query** method under the hood with "LET $@parameter = @value;"

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#let

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/let

        :param name: name for the variable (without $ sign!)
        :param value: value for the variable
        :return: result of request
        """
        try:
            data = json.dumps(value)
        except RecursionError as e:
            logger.error("Cant serialize object, too many nested levels")
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
        logger.info("Operation: LET. Name: %s, Value: %s", crop_data(name), crop_data(str(value)))
        return self.query(f"LET ${name} = {data};")

    @connected
    def unset(self, name: str) -> SurrealResult:
        """
        This method unsets value, which was previously stored

        Http transport use **query** method under the hood with "LET $@parameter = @value;"

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#unset

        :param name: name for the variable (without $ sign!)
        :return: result of request
        """
        logger.info("Operation: UNSET. Variable name: %s", crop_data(name))
        return self.query(f"REMOVE PARAM ${name};")

    @connected
    def use(self, namespace: str, database: str) -> SurrealResult:
        """
        This method specifies the namespace and database for the current connection

        Http transport use **query** method under the hood with "USE NS namespace DB database;"

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/use

        Example:
        http_connection.use("test", "test") # use test namespace and test database for this connection

        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :return: result of request
        """
        logger.info("Operation: USE. Namespace: %s, database %s", crop_data(namespace), crop_data(database))
        result = self.query(f"USE NS {namespace} DB {database};")
        if not result.is_error():
            # if USE was OK we need to store new data (ns and db)
            self._db_params = {"NS": namespace, "DB": database}
            self._http_client = HttpClient(self._url, headers=self._db_params, credentials=self._credentials,
                                           timeout=self._timeout)
        return result

    def patch(self, table_name: str, data: Dict, record_id: Optional[str] = None, return_diff: bool = False):
        """
        Http transport can not use patch operation, you can use websocket transport for that, or methods like
        **merge** and **update** for same results

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not use PATCH operation, use websocket transport or methods like merge and update"
        logger.error(message)
        raise CompatibilityError(message)

    def live(self, table_name, callback, return_diff: bool = False):
        """
        Http transport can not use live queries, you should use websocket transport for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not use live queries, use websockets instead"
        logger.error(message)
        raise CompatibilityError(message)

    def kill(self, live_query_id: str):
        """
        Http transport can not use KILL operation, you should use websocket transport for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not use kill, use websockets instead"
        logger.error(message)
        raise CompatibilityError(message)

    def insert(self, table_name: str, data: Dict):
        """
        Http transport can not insert more than 1 one record, you should use **create** method for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not insert more than 1 one record, you should use create method for that"
        logger.error(message)
        raise CompatibilityError(message)

    def authenticate(self, token: str):
        """
        Http transport can not use authenticate, you should use websocket for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not authenticate, you should use websocket for that"
        logger.error(message)
        raise CompatibilityError(message)

    def invalidate(self):
        """
        Http transport can not use invalidate, you should use websocket for that

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not invalidate, you should use websocket for that"
        logger.error(message)
        raise CompatibilityError(message)

    def info(self):
        """
        Http transport can not use INFO, you should use websocket or **query** method

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :raise CompatibilityError: on any use
        """
        message = "Http transport can not use INFO, you should use query method or websocket for that"
        logger.error(message)
        raise CompatibilityError(message)

    def _simple_get(self, endpoint: str) -> Tuple[int, str]:
        with self._http_client.get(endpoint) as resp:
            status, text = resp.status, resp.read().decode(ENCODING)
            _body = "is empty" if not text else crop_data(text)
            logger.info("Response from /%s, status_code: %s, body: %s", endpoint, status, _body)
            return status, text

    def _simple_request(self, method, endpoint: str, data: Union[Dict, str], not_json: bool = False) -> Tuple[int, str]:
        with self._http_client.request(method, data, endpoint, not_json) as resp:
            status, text = resp.status, resp.read().decode(ENCODING)
            _body = "is empty" if not text else crop_data(text)
            logger.info("Response from /%s, status_code: %s, body %s", endpoint, status, _body)
            return status, text


def raise_if_not_http_ok(result: Tuple[int, str]) -> str:
    """
    Helper for methods which need only success responses

    :param result: pair of status_code and text from the http response
    :return: text of the response if status is 200, else raise
    :raise HttpConnectionError: if status code is not 200
    """
    status, text = result
    if status != HTTP_OK:
        logger.error("Status code is %s, check your data! Info %s", status, text)
        text = text.replace(',', ',\n')
        raise HttpConnectionError(f"Status code is {status}, check your data! Info:\n{text}")
    return text



import dataclasses
import json
import re
import uuid
from typing import Union, List, Dict, Optional

from src.surrealist.errors import TooManyNestedLevelsError

ENCODING = "UTF-8"
OK = "OK"
HTTP_OK = 200  # status code for success
DEFAULT_TIMEOUT = 5  # timeout in seconds for basic operations
DATA_LENGTH_FOR_LOGS = 300  # size of data in logs, data will be cropped if bigger than that


def _set_length(length: int):
    global DATA_LENGTH_FOR_LOGS
    DATA_LENGTH_FOR_LOGS = length


def get_uuid() -> str:
    """
    Helper to generate uuid for records or testing purposes

    :return: uuid string representation
    """
    return str(uuid.uuid4())


def mask_pass(text: str) -> str:
    """
    Mask the passwords in logs. Replace all 'pass':'any_pass' in logs to 'pass':'******'. Works both with ' and "
    :param text: text before put it to logs
    :return: text without visible passwords
    """
    return re.sub(r"(?ms)(?<=['\"]pass['\"]: ['\"]).*?(?=['\"])", '******', text)


def crop_data(data: Union[str, bytes], is_str: bool = True) -> Union[str, bytes]:
    """
    Crop the data to maximum size, no actions if data is smaller, work with str and bytes in logs

    :param data: str or bytes of data to put it in logs
    :param is_str: flag to use correct operation
    :return: data of same type, but cropped if it bigger, than DATA_LENGTH_FOR_LOGS
    """
    if len(data) > DATA_LENGTH_FOR_LOGS:
        return f"{data[:DATA_LENGTH_FOR_LOGS]}..." if is_str else data[:DATA_LENGTH_FOR_LOGS] + b'...'
    return data


@dataclasses.dataclass
class SurrealResult:
    """
    Represent common object for all possible responses of SurrealDB
    """
    id: Optional[Union[int, str]] = None
    error: Optional[Dict] = None
    result: Optional[Union[str, Dict, List]] = None
    code: Optional[int] = None
    token: Optional[str] = None
    status: Optional[str] = OK
    time: str = ''

    def is_error(self):
        """
        Method to know the error was returned

        :return: True if response contains error, False otherwise
        """
        return self.error is not None


def to_result(content: Union[str, Dict]) -> SurrealResult:
    """
    Converts str or dict response of SurrealDB to common object for convenient use

    :param content: response from SurrealDB
    :return: Result object
    """
    if isinstance(content, str):
        try:
            content = json.loads(content)
        except RecursionError as exc:
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from exc
    if isinstance(content, List):
        content = content[0]
    if 'details' in content:
        content['result'] = content['details']
        del content['details']
    return SurrealResult(**content)



import json
import threading
import time
from json import JSONDecodeError
from logging import getLogger
from threading import Lock
from typing import Dict, Callable, Optional

import websocket

from src.surrealist.errors import WebSocketConnectionClosedError, TooManyNestedLevelsError
from src.surrealist.utils import to_result, SurrealResult, DEFAULT_TIMEOUT, get_uuid, crop_data, mask_pass

logger = getLogger("websocket_client")


class WebSocketClient:
    """
    Blocking thread-safe client to work with websockets, always wait a response for message, comparing by id (uuid).
    Every client creates at least 2 threads (in and out)
    """

    def __init__(self, base_url: str, timeout: int = DEFAULT_TIMEOUT):
        self._ws = None
        self._connected = None
        self._timeout = timeout
        self._base_url = base_url
        thread = threading.Thread(target=self.run, daemon=True)
        thread.start()
        logger.debug("Connecting to %s", base_url)
        self._raise_on_wait(lambda: self._connected is True, timeout=timeout,
                            error_text=f"Not connected to {self._base_url}")
        self.lock = Lock()
        self._callbacks = {}
        self._messages = {}
        logger.debug("Connected to %s, timeout is %s seconds", base_url, timeout)

    def on_message(self, _ws, message: str):
        """
        Called on message received from the websocket connection.

        :param _ws: connection object
        :param message: string message
        :return: None
        """
        logger.debug("Get message %s", crop_data(message))
        try:
            mess = json.loads(message)
        except JSONDecodeError as je:
            # Should never happen, all messages via json
            logger.error("Got non-json response %s", crop_data(message), exc_info=True)
            raise ValueError(f"Got non-json response! {message}") from je
        except RecursionError as e:
            logger.error("Cant deserialize object, too many nested levels")
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
        if "id" in mess:
            id_ = mess["id"]
            self._messages[id_] = mess
        else:
            # no id at top level = live query received
            if 'result' in mess:
                live_id = mess['result']['id']
                callback = self._callbacks.get(live_id)
                if callback:
                    logger.debug("Use callback for %s", live_id)
                    callback(mess)
                else:
                    logger.warning("Got message, but no callback to work with. Message: %s", mess)
            else:
                logger.warning("Got unexpected message without id and result:  %s", mess)

    def on_error(self, _ws, err: Exception):
        logger.error("Websocket connection gets an error %s", err)

    def is_connected(self) -> bool:
        """
        Shows is websocket client is connected to SurrealDB

        :return: True if connected, False otherwise
        """
        return bool(self._connected)

    def on_open(self, _ws):
        self._connected = True

    def on_close(self, *_ignore):
        self._connected = False
        logger.debug("Close connection to %s", self._base_url)

    def run(self):
        """
        Constantly waiting for messages on websocket connection, runs in separate thread

        :return: None
        """
        self._ws = websocket.WebSocketApp(self._base_url, on_open=self.on_open, on_message=self.on_message,
                                          on_error=self.on_error, on_close=self.on_close)
        self._ws.run_forever(skip_utf8_validation=True)

    def send(self, data: Dict, callback: Optional[Callable] = None) -> SurrealResult:
        """
        Method to send messages to SurrealDB, blocks until gets a response or raise on timeout

        :param data: dict with request parameters
        :param callback: function to call on live query, it is set only for live method
        :return: result of the request
        :raise TimeoutError: if no response and time is over
        :raise WebSocketConnectionClosed: if connection was closed while waiting
        """
        id_ = get_uuid()
        data = {"id": id_, **data}
        try:
            data_string = json.dumps(data, ensure_ascii=False)
        except RecursionError as e:
            logger.error("Cant serialize object, too many nested levels")
            raise TooManyNestedLevelsError("Cant serialize object, too many nested levels") from e
        logger.debug("Send data: %s", crop_data(mask_pass(data_string)))
        self._ws.send(data_string)
        res = self._get_by_id(id_)
        if data['method'] in ('live', 'kill'):
            if 'error' not in res:
                # now we know live or kill was successful, so now we need to manage callbacks
                self._on_success(data, callback, res)
        return to_result(res)

    def _on_success(self, data: Dict, callback: Callable, result: Dict):
        if data['method'] == 'kill':
            logger.debug("Delete callback for %s", data['params'][0])
            self._callbacks[data['params'][0]] = None
        if data['method'] == 'live':
            logger.debug("Set callback for %s", result['result'])
            self._callbacks[result['result']] = callback

    def _get_by_id(self, id_) -> Dict:
        self._raise_on_wait(lambda: id_ in self._messages, timeout=self._timeout,
                            error_text=f"No messages with id {id_} received in {self._timeout} seconds")
        with self.lock:
            result = self._messages.pop(id_, None)
        if result is None:
            # Should never happen!
            logger.error("Dict returns None on thread-safe pop, id %s, messages %s", id_, self._messages)
            raise ValueError(f"Dict returns None on thread-safe pop, {id_=}, {self._messages=}")
        return result

    def _wait_until(self, predicate, timeout, period=0.1):
        must_end = time.time() + timeout
        while time.time() < must_end:
            if self._connected is False:
                return False, "CLOSED"
            if predicate():
                return True, None
            time.sleep(period)
        return False, "TIME"

    def _raise_on_wait(self, predicate, timeout, error_text):
        """
        Method to wait some condition or raise on timeout

        :param predicate: function to call and check condition
        :param timeout: time in seconds to wait until condition
        :param error_text: custom error message on fail
        :raise TimeoutError: if condition not met until time
        :raise WebSocketConnectionClosed: if connection was closed while waiting
        """
        result = self._wait_until(predicate, timeout)
        if result == (False, "TIME"):
            logger.error("Time exceeded: %s seconds. Error: %s", timeout, error_text)
            raise TimeoutError(f"Time exceeded: {timeout} seconds. Error: {error_text}")
        if result == (False, "CLOSED"):
            logger.error("Connection %s closed while client waits on it", self._base_url)
            raise WebSocketConnectionClosedError("Connection closed while client waits on it")

    def close(self):
        self._connected = False
        self._ws.close()
        del self._ws
        self._messages.clear()
        self._callbacks.clear()
        logger.debug("Client is closed connection to %s", self._base_url)
